#include "Application.h"
#include "Benchmarker.h"
#include "InputManager.h"
#include "Logger.h"

#include "OgreEntity.h"
#include "OgreMaterial.h"
#include "OgreMaterialManager.h"
#include "OgreOverlay.h"
#include "OgreOverlayElement.h"
#include "OgreOverlayManager.h"
#include "OgrePass.h"
#include "OgreRoot.h"
#include "OgreSceneManager.h"
#include "OgreSceneNode.h"
#include "OgreTechnique.h"
#include <sstream>


Application::Application():
	mNumObjects(1),
	mMeshesIndex(0),
	mMaterialIndex(19),
	mDefaultWallFrequencies(2.99f, 7.99f, 2.99f),
	mWallFrequencies(mDefaultWallFrequencies),
	mDefaultLightThreshold(0.6f),
	mLightThreshold(mDefaultLightThreshold),
	mDefaultAlphaPlaneDistance(0.1f),
	mAlphaPlaneDistance(mDefaultAlphaPlaneDistance),
	mDefaultWallDisplacement(1.0f),
	mWallDisplacement(mDefaultWallDisplacement),
	mWallFrequenciesString("wallFrequencies"),
	mLightThresholdString("lightThreshold"),
	mAlphaPlaneDistanceString("alphaPlaneDistance"),
	mWallDisplacementString("displacementStrengths"),
	mCameraMode(CM_AUTOMATIC),
	mCameraAnimationPhase(0.0f),
	mBenchmarker(NULL),
	mInfoState(IS_SMALL)
{	
	mMeshes.push_back(MeshName("Box.mesh"				, "Box"							, Ogre::Vector3( 0.5f, 0.5f,  -0.5f),  2.00f));
	mMeshes.push_back(MeshName("Cylinder.mesh"			, "Cylinder"					, Ogre::Vector3( 0.0f, 0.5f,   0.0f),  1.90f));
	mMeshes.push_back(MeshName("Sphere.mesh"			, "Sphere"						, Ogre::Vector3( 0.0f, 0.0f,   0.0f),  1.40f));
	mMeshes.push_back(MeshName("ManyBuildings.mesh"		, "Many buildings in one mesh"	, Ogre::Vector3(15.5f, 1.5f, -15.5f), 22.10f));
	mMeshes.push_back(MeshName("PolygonedBuilding.mesh"	, "Not interior mapped building", Ogre::Vector3( 0.5f, 1.5f,  -0.5f),  3.65f));
	mMeshes.push_back(MeshName("BoxHigh.mesh"			, "Interior mapped building"	, Ogre::Vector3( 0.5f, 1.5f,  -0.5f),  3.65f));



	pushMaterial("IM_Ceiling"																	, "Interior ceilings only");
	pushMaterial("IM_CeilingFloor"																, "Interior ceilings and floors only");
	pushMaterial("Reflector"																	, "Cube reflections only");
	pushMaterial("IM_DiffuseCube"																, "Diffuse map and reflections and no interior mapping");
	pushMaterial("IM_DiffuseCube_Ceiling"														, "Diffuse map, reflections and interior ceilings");
	pushMaterial("IM_DiffuseCube_CeilingFloor"													, "Diffuse map, reflections and interior ceilings and floors");
	pushMaterial("IM_UntexturedRooms"															, "Untextured rooms");
	pushMaterial("IM_DiffuseCube_UntexturedRooms"												, "Diffuse map, reflections and untextured rooms");
	pushMaterial("IM_UniformTexturedRooms"														, "Uniformly textured rooms");
	pushMaterial("IM_DiffuseCube_UniformTexturedRooms"											, "Diffuse map, reflections and uniformly textured rooms");
	pushMaterial("IM_CeilingTexturedRooms"														, "Ceiling textured rooms");
	pushMaterial("IM_DiffuseCube_CeilingTexturedRooms"											, "Diffuse map, reflections and ceiling textured rooms");
	pushMaterial("IM_FullyTexturedRooms"														, "Fully textured rooms (256x256)");
	pushMaterial("IM_DiffuseCube_FullyTexturedRooms"											, "Diffuse map, reflections and fully textured rooms (256x256)");
	pushMaterial("IM_DiffuseCube_FullyTexturedRooms64"											, "Diffuse map, reflections and fully textured rooms (64x64)");
	pushMaterial("IM_DiffuseCube_FullyTexturedRooms16"											, "Diffuse map, reflections and fully textured rooms (16x16)");
	pushMaterial("IM_DiffuseCube_FullyTexturedRooms_EarlyZOut"									, "Diffuse map, reflections and fully textured rooms (256x256); early z-out");
	pushMaterial("IM_DiffuseCube_FullyTexturedRooms_OtherTextures"								, "Diffuse map, reflections and fully textured rooms, with other textures");
	pushMaterial("IM_VariationTexturedRooms"													, "Fully textured rooms with texture randomisation");
	pushMaterial("IM_DiffuseCube_VariationTexturedRooms"										, "Diffuse map, reflections and Fully textured rooms with texture randomisation");
	pushMaterial("IM_FullyTexturedRooms_LightVariation"											, "Fully textured rooms with variations in lighting");
	pushMaterial("IM_DiffuseCube_FullyTexturedRooms_LightVariation"								, "Diffuse map, reflections and fully textured rooms with variations in lighting");
	pushMaterial("IM_FullyTexturedRooms_AlphaPlane"												, "Fully textured rooms with alpha planes");
	pushMaterial("IM_DiffuseCube_FullyTexturedRooms_AlphaPlane"									, "Diffuse map, reflections and fully textured rooms with alpha planes");
	pushMaterial("IM_FullyTexturedRooms_AlphaPlane_LightVariation"								, "Fully textured rooms with alpha planes and variations in lighting");
	pushMaterial("IM_DiffuseCube_FullyTexturedRooms_AlphaPlane_LightVariation"					, "Diffuse map, reflections and fully textured rooms with alpha planes and variations in lighting");
	pushMaterial("IM_DiffuseCube_FullyTexturedRooms_AlphaPlane_LightVariation_OtherTextures"	, "Diffuse map, reflections and fully textured rooms with alpha planes and variations in lighting, with other textures");
	pushMaterial("IM_FullyTexturedRooms_DisplacedWalls"											, "Fully textured rooms with displaced walls");
	pushMaterial("IM_DiffuseCube_FullyTexturedRooms_DisplacedWalls"								, "Diffuse map, reflections and fully textured rooms with displaced walls");
	

	/*
	//usefull order for capturing video from the demo application
	pushMaterial("IM_DiffuseCube_FullyTexturedRooms_OtherTextures"					, "Diffuse map, reflections and fully textured rooms, with other textures");
	//pushMaterial("IM_Ceiling"														, "Interior ceilings only");
	//pushMaterial("IM_CeilingFloor"													, "Interior ceilings and floors only");
	//pushMaterial("Reflector"														, "Cube reflections only");
	//pushMaterial("IM_DiffuseCube"													, "Diffuse map and reflections and no interior mapping");
	//pushMaterial("IM_DiffuseCube_Ceiling"											, "Diffuse map, reflections and interior ceilings");
	//pushMaterial("IM_DiffuseCube_CeilingFloor"										, "Diffuse map, reflections and interior ceilings and floors");
	pushMaterial("IM_UntexturedRooms"												, "Untextured rooms");
	//pushMaterial("IM_DiffuseCube_UntexturedRooms"									, "Diffuse map, reflections and untextured rooms");
	//pushMaterial("IM_UniformTexturedRooms"											, "Uniformly textured rooms");
	//pushMaterial("IM_Diffuse_UniformTexturedRooms"									, "Diffuse map and uniformly textured rooms");
	//pushMaterial("IM_DiffuseCube_UniformTexturedRooms"								, "Diffuse map, reflections and uniformly textured rooms");
	pushMaterial("IM_CeilingTexturedRooms"											, "Ceiling textured rooms");
	//pushMaterial("IM_Diffuse_CeilingTexturedRooms"									, "Diffuse map and ceiling textured rooms");
	pushMaterial("IM_DiffuseCube_CeilingTexturedRooms"								, "Diffuse map, reflections and ceiling textured rooms");
	pushMaterial("IM_FullyTexturedRooms"											, "Fully textured rooms (256x256)");
	pushMaterial("IM_DiffuseCube_FullyTexturedRooms"								, "Diffuse map, reflections and fully textured rooms (256x256)");
	//pushMaterial("IM_DiffuseCube_FullyTexturedRooms64"								, "Diffuse map, reflections and fully textured rooms (64x64)");
	//pushMaterial("IM_DiffuseCube_FullyTexturedRooms16"								, "Diffuse map, reflections and fully textured rooms (16x16)");
	//pushMaterial("IM_DiffuseCube_FullyTexturedRooms_EarlyZOut"						, "Diffuse map, reflections and fully textured rooms (256x256); early z-out");
	//pushMaterial("IM_FullyTexturedRooms_LightVariation"								, "Fully textured rooms with variations in lighting");
	//pushMaterial("IM_DiffuseCube_FullyTexturedRooms_LightVariation"					, "Diffuse map, reflections and fully textured rooms with variations in lighting");
	//pushMaterial("IM_DiffuseCube_FullyTexturedRooms_LightVariation_OtherTextures"	, "Diffuse map, reflections and fully textured rooms with variations in lighting, with other textures");
	pushMaterial("IM_FullyTexturedRooms_AlphaPlane"									, "Fully textured rooms with alpha planes");
	pushMaterial("IM_DiffuseCube_FullyTexturedRooms_AlphaPlane"						, "Diffuse map, reflections and fully textured rooms with alpha planes");
	pushMaterial("IM_FullyTexturedRooms_AlphaPlane_LightVariation"					, "Fully textured rooms with alpha planes and variations in lighting");
	pushMaterial("IM_DiffuseCube_FullyTexturedRooms_AlphaPlane_LightVariation"		, "Diffuse map, reflections and fully textured rooms with alpha planes and variations in lighting");
	pushMaterial("IM_FullyTexturedRooms_DisplacedWalls"								, "Fully textured rooms with displaced walls");
	pushMaterial("IM_DiffuseCube_FullyTexturedRooms_DisplacedWalls"					, "Diffuse map, reflections and fully textured rooms with displaced walls");
	*/
}


Application::~Application()
{
	if (mBenchmarker)
	{
		delete mBenchmarker;
	}
	destroyObjects();
	InputManager::get()->removeKeyListener(this);
}


void Application::createScene()
{
	InputManager::get()->addKeyListener(this);

	createObjects();

	Ogre::OverlayManager::getSingleton().getByName("HUD")->show();
	updateOverlayText();
}


void Application::createFrameListener()
{
	ExampleApplication::createFrameListener();
	mRoot->addFrameListener(this);
}


void Application::keyPressed(const OIS::KeyEvent& e)
{
	int key = e.key;
	switch (key)
	{
	case OIS::KC_F1:
		mInfoState = InfoState((mInfoState + 1) % 3);
		break;
	case OIS::KC_COMMA:
		switchMaterial(false);
		break;
	case OIS::KC_PERIOD:
		switchMaterial(true);
		break;
	case OIS::KC_N:
		switchMesh(false);
		break;
	case OIS::KC_M:
		switchMesh(true);
		break;
	case OIS::KC_V:
		addToNumObjects(-1);
		break;
	case OIS::KC_B:
		addToNumObjects(1);
		break;	
	case OIS::KC_F5:
		if (!mBenchmarker)
		{
			mBenchmarker = new Benchmarker(this, mCamera);
		}
	}

	if (mCameraMode == CM_FREE)
	{
		if (e.key == OIS::KC_C)
		{
			mCameraMode = CM_AUTOMATIC;
		}
	}
	else
	{
		if (e.key == OIS::KC_C
			|| e.key == OIS::KC_W
			|| e.key == OIS::KC_A
			|| e.key == OIS::KC_S
			|| e.key == OIS::KC_D
			|| e.key == OIS::KC_PGDOWN
			|| e.key == OIS::KC_PGUP
			|| e.key == OIS::KC_LEFT
			|| e.key == OIS::KC_RIGHT
			|| e.key == OIS::KC_UP
			|| e.key == OIS::KC_DOWN)
		{
			mCameraMode = CM_FREE;
		}
	}
}


bool Application::frameStarted(const Ogre::FrameEvent& evt)
{
	float time = evt.timeSinceLastFrame;

	if (mBenchmarker)
	{
		mBenchmarker->update(time);
		if (mBenchmarker->isFinished())
		{
			delete mBenchmarker;
			mBenchmarker = NULL;
		}
	}
	else
	{
		updateKeysDown(time);
		updateCameraAnimation(time);
		updateOverlayText();
	}
	updateUniformParameters();

	return true;
}


size_t Application::getNumMaterials() const
{
	return mMaterials.size();
}


std::string Application::getCurrentMaterialName() const
{
	return mMaterials[mMaterialIndex].name;
}


void Application::switchMaterial(bool nextMaterial)
{
	if (currentMeshHasSwappableMaterial())
	{
		size_t newIndex;
		if (nextMaterial)
		{
			newIndex = (mMaterialIndex + 1) % mMaterials.size();
		}
		else
		{
			newIndex = (mMaterialIndex + mMaterials.size() - 1) % mMaterials.size();
		}
		switchToMaterial(newIndex);
	}
}


void Application::switchToMaterial(size_t materialIndex)
{
	mMaterialIndex = materialIndex;
	if (mMaterialIndex >= mMaterials.size())
	{
		mMaterialIndex = mMaterials.size() - 1;
	}
	setEntitiesMaterial();
	mWallFrequencies = mDefaultWallFrequencies;
	mLightThreshold = mDefaultLightThreshold;
	mAlphaPlaneDistance = mDefaultAlphaPlaneDistance;
	mWallDisplacement = mDefaultWallDisplacement;
}


void Application::addToNumObjects(int amount)
{
	setNumObjects(mNumObjects + amount);
}


void Application::setNumObjects(int numObjects)
{
	mNumObjects = numObjects;
	if (mNumObjects <= 0)
	{
		mNumObjects = 1;
	}
	createObjects();
}


size_t Application::getNumMeshes() const
{
	return mMeshes.size();
}


void Application::switchMesh(bool nextMesh)
{
	size_t newIndex;
	if (nextMesh)
	{
		newIndex = (mMeshesIndex + 1) % mMeshes.size();
	}
	else
	{
		newIndex = (mMeshesIndex + mMeshes.size() - 1) % mMeshes.size();
	}
	switchToMesh(newIndex);
}


void Application::switchToMesh(size_t meshIndex)
{
	mMeshesIndex = meshIndex;
	if (mMeshesIndex >= mMeshes.size())
	{
		mMeshesIndex = mMeshes.size() - 1;
	}
	createObjects();
}


void Application::setWallFrequences(const Ogre::Vector3& wallFrequencies)
{
	mWallFrequencies = wallFrequencies;
}


void Application::setOverlayTexts(const std::string& text0, const std::string& text1) const
{
	Ogre::OverlayManager& overlayManager = Ogre::OverlayManager::getSingleton();
	overlayManager.getOverlayElement("HUD/Materials")->setCaption(text0);
	overlayManager.getOverlayElement("HUD/Buttons")->setCaption(text1);
}


void Application::pushMaterial(const std::string& materialName,
							   const std::string& description)
{
	mMaterials.push_back(MaterialName(materialName, description));
}


void Application::destroyObjects()
{
	size_t entitiesSize = mEntities.size();
	for (size_t i = 0; i < entitiesSize; ++i)
	{
		if (mEntities[i]->isAttached())
		{
			mSceneNodes[i]->detachObject(mEntities[i]);
		}
		mSceneMgr->destroyEntity(mEntities[i]);
	}
	mEntities.clear();

	size_t nodesSize = mSceneNodes.size();
	for (size_t i = 0; i < nodesSize; ++i)
	{
		mSceneNodes[i]->getParentSceneNode()->removeAndDestroyChild(mSceneNodes[i]->getName());
	}
	mSceneNodes.clear();
}


void Application::createObjects()
{
	destroyObjects();

	Ogre::SceneNode* rootSceneNode = mSceneMgr->getRootSceneNode();

	for (int x = 0; x < mNumObjects; ++x)
	{
		for (int z = 0; z < mNumObjects; ++z)
		{
			std::stringstream name;
			name << "Object_" << x << "_" << z << "_";
			Ogre::SceneNode* node = rootSceneNode->createChildSceneNode(name.str() + "SceneNode");
			Ogre::Entity* entity = mSceneMgr->createEntity(name.str() + "Entity", mMeshes[mMeshesIndex].name);
			node->setPosition(Ogre::Vector3(x * 2, 0, -z * 2));
			node->attachObject(entity);
			mSceneNodes.push_back(node);
			mEntities.push_back(entity);
		}
	}

	setEntitiesMaterial();
}


void Application::rotateObjects(Ogre::Degree amount)
{
	size_t size = mSceneNodes.size();
	for (size_t i = 0; i < size; ++i)
	{
		mSceneNodes[i]->roll(amount);
	}
}


bool Application::currentMeshHasSwappableMaterial() const
{
	return meshHasSwappableMaterial(mMeshesIndex);
}


bool Application::meshHasSwappableMaterial(size_t meshIndex) const
{
	std::string meshName = mMeshes[meshIndex].name;
	return meshName != "PolygonedBuilding.mesh"
		&& meshName != "BoxHigh.mesh";
}


void Application::setEntitiesMaterial()
{
	//only change the material for swappable materials
	if (currentMeshHasSwappableMaterial())
	{
		size_t size = mEntities.size();
		for (size_t i = 0; i < size; ++i)
		{
				mEntities[i]->setMaterialName(mMaterials[mMaterialIndex].name);
		}
	}
}


Ogre::Pass* Application::getPass() const
{
	Ogre::Technique* technique =
		static_cast<Ogre::MaterialPtr>(Ogre::MaterialManager::getSingleton().getByName(mMaterials[mMaterialIndex].name))
		->getTechnique(0);

	Ogre::Pass* pass = technique->getPass(technique->getNumPasses() - 1);

	return pass;
}


void Application::updateKeysDown(float time)
{
	InputManager* inputManager = InputManager::get();

	//only change material parameters if the current object has a swappable material
	if (currentMeshHasSwappableMaterial())
	{
		if (inputManager->isKeyDown(OIS::KC_T)) mAlphaPlaneDistance += time * 0.5f;
		if (inputManager->isKeyDown(OIS::KC_G)) mAlphaPlaneDistance -= time * 0.5f;

		if (inputManager->isKeyDown(OIS::KC_2)) mWallDisplacement += time * 0.75f;
		if (inputManager->isKeyDown(OIS::KC_1)) mWallDisplacement -= time * 0.75f;
		mWallDisplacement = std::max(0.0f, std::min(2.0f, mWallDisplacement));

		if (inputManager->isKeyDown(OIS::KC_Y)) mLightThreshold += time * 0.25f;
		if (inputManager->isKeyDown(OIS::KC_H)) mLightThreshold -= time * 0.25f;
		mLightThreshold = std::max(0.0f, std::min(1.0f, mLightThreshold));

		if (inputManager->isKeyDown(OIS::KC_U)) mWallFrequencies.x *= pow(2.0f, time);
		if (inputManager->isKeyDown(OIS::KC_J)) mWallFrequencies.x *= pow(0.5f, time);
		if (inputManager->isKeyDown(OIS::KC_I)) mWallFrequencies.y *= pow(2.0f, time);
		if (inputManager->isKeyDown(OIS::KC_K)) mWallFrequencies.y *= pow(0.5f, time);
		if (inputManager->isKeyDown(OIS::KC_O)) mWallFrequencies.z *= pow(2.0f, time);
		if (inputManager->isKeyDown(OIS::KC_L)) mWallFrequencies.z *= pow(0.5f, time);
	}

	//handle rotating the entire mesh
	if (inputManager->isKeyDown(OIS::KC_LBRACKET)) rotateObjects(time * Ogre::Degree(90));
	if (inputManager->isKeyDown(OIS::KC_RBRACKET)) rotateObjects(time * Ogre::Degree(-90));
}


void Application::updateUniformParameters()
{
	Ogre::Pass* pass = getPass();
	if (hasParameter(mWallFrequenciesString))
	{
		pass->getFragmentProgramParameters()->setNamedConstant(mWallFrequenciesString, mWallFrequencies);
	}
	if (hasParameter(mLightThresholdString))
	{
		pass->getFragmentProgramParameters()->setNamedConstant(mLightThresholdString, mLightThreshold);
	}
	if (hasParameter(mAlphaPlaneDistanceString))
	{
		pass->getFragmentProgramParameters()->setNamedConstant(mAlphaPlaneDistanceString, mAlphaPlaneDistance);
	}
	if (hasParameter(mWallDisplacementString))
	{
		Ogre::Vector3 wallDisplacementVector(mWallDisplacement, 0, mWallDisplacement);
		pass->getFragmentProgramParameters()->setNamedConstant(mWallDisplacementString, wallDisplacementVector);
	}
}


void Application::updateCameraAnimation(float time)
{
	if (mCameraMode == CM_AUTOMATIC)
	{
		mCameraAnimationPhase += time;
		Ogre::Vector3 animationCentre = mMeshes[mMeshesIndex].lookAtCentre;
		Ogre::Vector3 localPosition(sin(0.25f * mCameraAnimationPhase),
									cos(0.11f * mCameraAnimationPhase) * 0.25f,
									cos(0.25f * mCameraAnimationPhase));
		mCamera->setPosition(animationCentre + mMeshes[mMeshesIndex].cameraDistance * localPosition);
		mCamera->lookAt(animationCentre);
	}
}


bool Application::hasParameter(const std::string& parameterName) const
{
	Ogre::Pass* pass = getPass();
	if (pass->hasFragmentProgram())
	{
		Ogre::GpuProgramParametersSharedPtr parameters = pass->getFragmentProgramParameters();
		if (!parameters.isNull() && parameters->hasNamedParameters())
		{
			Ogre::GpuConstantDefinitionMap map = parameters->getConstantDefinitions().map;
			return map.find(parameterName) != map.end();
		}
	}
	return false;
}


void Application::updateOverlayText() const
{
	std::stringstream text0, text1;
	if (mInfoState != IS_NONE)
	{
		text0 << "Interior Mapping version 12";
	}
	if (mInfoState == IS_FULL)
	{
		updateFullInformation(text0, text1);
	}
	else if (mInfoState == IS_SMALL)
	{
		updateSmallInformation(text0);
	}
	setOverlayTexts(text0.str(), text1.str());
}


void Application::updateFullInformation(std::stringstream& text0, std::stringstream& text1) const
{
	bool swappableMaterial = currentMeshHasSwappableMaterial();
	bool hasWallFrequencies = hasParameter(mWallFrequenciesString) && swappableMaterial;
	bool hasLightThreshold = hasParameter(mLightThresholdString) && swappableMaterial;
	bool hasAlphaPlaneDistance = hasParameter(mAlphaPlaneDistanceString) && swappableMaterial;
	bool hasWallDisplacement = hasParameter(mWallDisplacementString) && swappableMaterial;

	text1
		<< "Keys:"
		<< "\n  F1: hide information"
		<< "\n  F: show/hide framerate"
		<< "\n  N and M: switch the mesh"
		<< "\n  V and B: change the number of meshes"
		<< "\n  [ and ]: rotate the object"
		<< "\n  , (comma) and . (period): switch material"
		<< "\n  R: wireframe"
		<< "\n  Print Screen: save screenshot"
		<< "\n  F5: Start benchmark";

	text1 << "\n  C: ";
	if (mCameraMode == CM_FREE)
	{
		text1 << "switch to automatic camera";
	}
	else
	{
		text1 << "switch to free camera";
	}
	
	text1 << "\n";
	if (hasWallFrequencies)
	{
		text1 << "  UJIKOL: change wall frequencies";
	}

	text1 << "\n";
	if (hasLightThreshold)
	{
		text1	<< "  YH: change light threshold";
	}

	text1 << "\n";
	if (hasAlphaPlaneDistance)
	{
		text1	<< "  TG: change alpha plane distance";
	}

	text1 << "\n";
	if (hasWallDisplacement)
	{
		text1	<< "  12: change wall displacement";
	}

	//write a list of all meshes
	text1 << "\n\nMeshes:";
	size_t meshesSize = mMeshes.size();
	for (size_t i = 0; i < meshesSize; ++i)
	{
		if (i == mMeshesIndex)
		{
			text1 << "\n >" << toUpperCase(mMeshes[i].description);
		}
		else
		{
			text1 << "\n  " << mMeshes[i].description;
		}
	}
	text1 << "\n\nNumber of meshes: " << mNumObjects * mNumObjects;

	if (swappableMaterial)
	{
		//write a list of all materials
		text0 << "\n\nMaterials:";
		size_t matSize = mMaterials.size();
		for (size_t i = 0; i < matSize; ++i)
		{
			if (i == mMaterialIndex)
			{
				text0 << "\n >" << toUpperCase(mMaterials[i].description);
			}
			else
			{
				text0 << "\n  " << mMaterials[i].description;
			}
		}
	}

	if (hasWallFrequencies)
	{
		text0 << "\n\nWall frequencies:"
			<< "\n  YZ-walls: " << mWallFrequencies.x
			<< "\n  XZ-walls: " << mWallFrequencies.y
			<< "\n  XY-walls: " << mWallFrequencies.z;
	}

	if (hasLightThreshold)
	{
		text0 << "\nLight threshold: " << mLightThreshold;	
	}

	if (hasAlphaPlaneDistance)
	{
		text0 << "\nAlpha plane distance: " << mAlphaPlaneDistance;
	}

	if (hasWallDisplacement)
	{
		text0 << "\nWall displacement: " << mWallDisplacement;
	}
}


void Application::updateSmallInformation(std::stringstream& text0) const
{
	text0 << "\nCurrent material: ";
	std::string meshName = mMeshes[mMeshesIndex].name;
	if (meshName == "PolygonedBuilding.mesh")
	{
		text0 << "Polygonized building, no Interior Mapping";
	}
	else if (meshName == "BoxHigh.mesh")
	{
		text0 << "Diffuse map, reflections and fully textured rooms (256x256); early z-out";
	}
	else
	{
		text0 << mMaterials[mMaterialIndex].description;
	}
	text0 << "\nPress 'F1' for more information.";
}


std::string Application::toUpperCase(std::string text)
{
	int diff = 'A' - 'a';
	size_t size = text.size();
	for (size_t i = 0; i < size; ++i)
	{
		if (text[i] >= 'a' && text[i] <= 'z')
		{
			text[i] += diff;
		}
	}
	return text;
}